var _client, _connection, _socket, _buffer = "";
var send;
var connected = false;

if (typeof module !== 'undefined')
{
  EventEmitter = require("events");
  emitter = new EventEmitter();
  dispatchEvent = (id, arg) => emitter.emit(id, arg);
} else
{
  dispatchEvent = (id, data) => document.body.dispatchEvent(new CustomEvent(id, {detail:data}));
}

//sleeper = (ms) => function(x) { return new Promise(resolve => setTimeout(() => resolve(x), ms)); };

// nodejs 
function connectNode(url, WebSocketClient)
{
  return new Promise((resolve, reject) => {
    _client = new WebSocketClient();
     
    _client.on('connectFailed', function(error) {
        console.log('Connect Error: ' + error.toString());
        reject();
    });
     
    _client.on('connect', function(connection) {
        _connection = connection;
        send = (msg) => {
          if (_handlersLog && _handlersLog.send)
            _handlersLog.send(msg);

          connection.sendUTF(msg);
        }

        console.log('WebSocket Client Connected');
	connected = true;
        dispatchEvent("sdkConnected");

        connection.on('error', function(error) {
            console.log("Connection Error: " + error.toString());
        });

        connection.on('close', function() {
            console.log('Connection Closed');
	    connected = false;
            dispatchEvent("sdkDisconnected");
        });

        connection.on('message', function(message) {
          _buffer += new String(message.utf8Data);
          var lines = _buffer.split("\n");
          while ( lines.length > 1 )
          {
            var line = lines.shift();

            if (_handlersLog && _handlersLog.receive)
              _handlersLog.receive(line);

            eval(line);
          }
          _buffer = lines.join("\n");
        });
        resolve();
    });

    _client.connect(url);
  });
}

function disconnectNode()
{
  _connection.close();
  return Promise.resolve();
}

// web
function connectWeb(url, client)
{
  return new Promise((resolve, reject) => {
    var host;
    if (url)
      host = url;
    else
      host = window.location.hostname || "localhost";

    url = "ws://"+host+":38950/index";

    var notified = false;

    try {
      _socket = new WebSocket(url);
    } catch (e)
    {
      reject();
      return;
    }

    _socket.onmessage = (msg) =>
    {
      var token1 = "rpcResolve" + _rpcToken;
      var token2 = "rpcNotify"; // + _rpcToken; // toto generuje app

      _buffer += new String(msg.data);
      var lines = _buffer.split("\n");
      while ( lines.length > 1 )
      {
        var line = lines.shift();
        if (_handlersLog && _handlersLog.receive)
          _handlersLog.receive(line);

        if (line.substr(0, 2) == "g." ||
            line.substr(0, token1.length) == token1 || 
            line.substr(0, token2.length) == token2)
        {
          setTimeout(line, 0);
        }
      }
      _buffer = lines.join("\n");
    }

    _socket.onopen = () => { 
      connected = true;
      console.log("Connected!");
      if (!notified)
      {
        dispatchEvent("sdkConnected");

        notified = true;
        resolve();
      }
    };

    _socket.onclose = () => { 
      if (typeof(livetimer) != "undefined")
      {
        clearTimeout(livetimer);
        delete livetimer;
      }
      connected = false;
      dispatchEvent("sdkDisconnected");

      if (!notified)
      {
        notified = true;
        reject();
      }
    };

    send = (msg) => 
    {  
      if (_handlersLog && _handlersLog.send) 
        _handlersLog.send(msg); 
      _socket.send(msg); 

      return true;
    }
  });
}

function disconnectWeb()
{
  _socket.close();
  return Promise.resolve();
}

function jsonToObject(json)
{                     
  if (typeof(json) != "Object")
    return json;

  var out = [];
  for (var i in json)
   out[i] = jsonToObject(json[i]);
}

// handlers
function stringify(json)
{
  json = JSON.stringify(json);
  json = json.replace(/\"([^(\")"]+)\":/g,"$1:");  //This will remove all the quotes
  return json; 
}

_handlers = null;
_handlersLog = null;

var rpcResolveHandlers = [];
var rpcTasks = [];

// external handlers
function rpcResolveInternal(data)
{
  if (rpcResolveHandlers.length > 0)
  {
    rpcResolveHandlers.shift()(data);
  } else {
    console.log("Unmatched response: " + JSON.stringify(data));
  }
  RpcDo();
}

function rpcNotify(data)
{
  console.log(data);
  dispatchEvent(data.id, data);
}

var _rpcResolveRoute;
var _rpcResolveJunctions;
var _rpcBusy = false;
var _rpcToken = Math.floor(Math.random()*1000);

eval("rpcResolve" + _rpcToken + " = rpcResolveInternal");

function rpcResolveRouteAddStart()
{
  _rpcResolveRoute = [];
}

function rpcResolveRouteAddPoints(points)
{
  for (var i=0; i<points.length; i++)
  {
    _rpcResolveRoute.push(points[i].lX);
    _rpcResolveRoute.push(points[i].lY);
  }
}

function rpcResolveRouteAddDone()
{
  rpcResolveInternal(_rpcResolveRoute);
  _rpcBusy = false; 
}

function rpcResolveRouteJunctionStart()
{
  _rpcResolveJunctions = [];
}

function rpcResolveRouteJunctionAdd(junction)
{
  _rpcResolveJunctions.push(junction);
}

function rpcResolveRouteJunctionDone()
{
  rpcResolveInternal(_rpcResolveJunctions);
  _rpcBusy = false;
}

eval("rpcResolve" + _rpcToken + "RouteAddStart = rpcResolveRouteAddStart");
eval("rpcResolve" + _rpcToken + "RouteAddPoints = rpcResolveRouteAddPoints");
eval("rpcResolve" + _rpcToken + "RouteAddDone = rpcResolveRouteAddDone");
eval("rpcResolve" + _rpcToken + "RouteJunctionStart = rpcResolveRouteJunctionStart");
eval("rpcResolve" + _rpcToken + "RouteJunctionAdd = rpcResolveRouteJunctionAdd");
eval("rpcResolve" + _rpcToken + "RouteJunctionDone = rpcResolveRouteJunctionDone");

var rpcInterval;
function RpcDo()
{ 
  if (!rpcInterval)
  {
    rpcInterval = setInterval(() =>
    {
      if (!_rpcBusy && rpcTasks.length)
      {
        if (rpcTasks[0].block)
        {
          rpcWillBeBusy = false;
          _rpcBusy = true;
          send(rpcTasks.shift().command);
          return;
        }
        send(rpcTasks.map(t=>t.command).join(""));      
        rpcTasks = [];
      }
    }, 100)
  }
}

Sdk = {
  on:(event, handler) => 
  {
    if (typeof module === 'undefined')
      document.body.addEventListener(event, (e) => handler(e.detail), false);
    else
      emitter.on(event, handler);
  },
  isConnected: () => connected,
  Connect:typeof module === 'undefined' ? connectWeb : connectNode,
  Disconnect:typeof module === 'undefined' ? disconnectWeb : disconnectNode,
  Send:(msg, block)=>{
    rpcTasks.push({command:msg, block:block});
    RpcDo();
    return Promise.resolve();
  },
  EnableEcho:() => {
    _handlersLog = { 
      send:(x) => console.log("< " + (x+"#").replace("\n#", "")), 
      receive:(x) => { if (1 || x.indexOf(".onlineio.") == -1) console.log("> " + x) }
    };
  },
  DisableEcho:() => {
    _handlersLog = { 
      send:(x) => 0, 
      receive:(x) => 0
    };
  },

  Ping:() => {
    return new Promise((resolve, reject) => {
      _pingTimeout = setTimeout(reject, 5000);
      rpcResolveHandlers.push((r) => { if (r==10) {clearTimeout(_pingTimeout); resolve(); return; } reject(); });
      Sdk.Send("_rpcResolve"+_rpcToken+"(5+5);\n");
    });
  },

  MapLoader: {
    GetCurrentMlm: () => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(CAuraWnd.CurrentMap());\n");
    }),
    GetMapsCount: () => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(GetMapCount());\n");
    }),
    GetMap: (i) => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(GetMapInfo("+i+"));\n");
    }),
  },

  MapView: {
    GetVisibleArea: () => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(GetVisibleArea());\n");
    }),
    SetTilt: (x) => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(C3DMapWnd.SetWantedTilt("+x+"));\n");
    }),
    SetRotation: (x) => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(C3DMapWnd.SetWantedRotation("+x+"));\n");
    }),
    SetZoom: (x) => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(C3DMapWnd.SetWantedDistance("+Math.pow(2, (20-x)*1.5).toFixed(0)+"));\n");
    }),
    SetPosition: (lat, lon) => new Promise((resolve) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(C3DMapWnd.SetWantedPosition(" + Math.floor(lon*100000) + "," + Math.floor(lat*100000) + "));\n");
    }),

    GetRealLevel: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(C3DMapWnd.GetRealLevel());\n");
      rpcResolveHandlers.push(resolve);
    })
  },

  Store: {
    GetServer: () => new Promise((resolve, reject) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(Store.GetServer());\n");
    }),
    Install: (mapId) => new Promise((resolve, reject) => {
      rpcResolveHandlers.push(resolve);
      Sdk.Send("_rpcResolve"+_rpcToken+"(Store.Install('"+mapId+"'));\n");
    }),
  },
  Position: {
    SetAll: (lat, lon, kmh, heading) => new Promise((resolve, reject) => {
      Sdk.Send(
        "GPS.SetSpeed("+Math.floor(kmh)+");\n" +
        "GPS.SetCourse("+Math.floor(heading)+");\n" +
        "_rpcResolve"+_rpcToken+"(GPS.SetPosition("+Math.floor(lon*100000) + "," + Math.floor(lat*100000) +"));\n");
      rpcResolveHandlers.push(resolve);
    }),
    SetSpeed: (kmh) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(GPS.SetSpeed("+Math.floor(kmh)+"));\n");
      rpcResolveHandlers.push(resolve);
    }),
    SetHeading: (heading) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(GPS.SetCourse("+Math.floor(heading)+"));\n");
      rpcResolveHandlers.push(resolve);
    }),
    SetPosition: (lat, lon) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(GPS.SetPosition("+Math.floor(lon*100000) + "," + Math.floor(lat*100000) +"));\n");
      rpcResolveHandlers.push(resolve);
    }),
    GetPosition: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(GPS.GetPosition());\n");
      rpcResolveHandlers.push((resp) => 
      { 
        if (typeof(resp.lY) == "undefined" || resp.lY == -999999999)
          reject();
        else
          resolve({lat:resp.lY/100000, lon:resp.lX/100000}); 
      });
    })
  },
  Route: {
    Get: () => new Promise((resolve, reject) => {
      Sdk.Send("Route.Get('rpcResolve"+_rpcToken+"');\n");
      rpcResolveHandlers.push(resolve);
    }),
    Set: (viaPoints) => new Promise((resolve, reject) => {
      if (viaPoints.length < 2)
        throw "wrong number of via points";
      var commands = [];
      commands.push("Route.SetNaviType('car');");
      for (var i=0; i<viaPoints.length; i++)
      {
        var point = Math.floor(viaPoints[i].lng * 100000) + "," + Math.floor(viaPoints[i].lat * 100000);
        if (i==0)
          commands.push("Route.SetStart(" + point + ");");
        else if (i==viaPoints.length-1)
          commands.push("Route.SetFinish(" + point + ");");
        else 
          commands.push("Route.SetVia(" + point + ");");
      }
      commands.push("_rpcResolve"+_rpcToken+"(Route.Calculate());");
      Sdk.Send(commands.join("\n") + "\n");
      rpcResolveHandlers.push(resolve);
    }),

  },

  Demonstrate: {
    Start: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Demonstrate.Start());\n");
      rpcResolveHandlers.push(resolve);
    }),
    Stop: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Demonstrate.Stop());\n");
      rpcResolveHandlers.push(resolve);
    }),
    SetSpeed: (speed) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Demonstrate.SetSpeed("+Math.floor(speed)+"));\n");
      rpcResolveHandlers.push(resolve);
    }),
    GetSpeed: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Demonstrate.GetSpeed());\n");
      rpcResolveHandlers.push(resolve);
    })
  },

  MapMode: {
    LockVehicle: (lock) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(MapMode.LockVehicle("+(lock?1:0)+"));\n");
      rpcResolveHandlers.push(resolve);
    })
  },

  Geocoder: {
    Forward: (lat, lon) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Geocoder.Forward("+lat.toFixed(5)+","+lon.toFixed(5)+"));\n");
      rpcResolveHandlers.push(resolve);
    })
  },

  MapReader: {
    GetRoads: (lat1, lon1, lat2, lon2) => new Promise((resolve, reject) => {
      lat1 = lat1.toFixed(5);
      lon1 = lon1.toFixed(5);
      lat2 = lat2.toFixed(5);
      lon2 = lon2.toFixed(5);
      Sdk.Send("_rpcResolve"+_rpcToken+"(MapReader.GetRoads("+lon1+","+lat1+","+lon2+","+lat2+"));\n");
      rpcResolveHandlers.push(resolve);
    }),
    GetRoad: (info) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(MapReader.GetRoad('"+info.iso+"',"+info.id+"));\n");
      rpcResolveHandlers.push(resolve);
    }),
    Reload: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(MapReader.Reload());\n");
      rpcResolveHandlers.push(resolve);
    })
  },
  Navigation: {
    GetJunctions: () => new Promise((resolve, reject) => {
      Sdk.Send("Route.GetJunctions('rpcResolve"+_rpcToken+"');\n", true);
      rpcResolveHandlers.push(resolve);
    }),
  },

  Traffic: {
    Reset: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Traffic.Reset());\n");
      rpcResolveHandlers.push(resolve);
    }),

  },

  Skin: {
    Get: (id, attribute) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Xml.Get('app/"+id+"/"+attribute+"'));\n");
      rpcResolveHandlers.push(resolve);
    }),
    Set: (id, attribute, value) => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Xml.Set('app/"+id+"/"+attribute+"', '"+value+"'));\n");
      rpcResolveHandlers.push(resolve);
    }),
    Reload: () => new Promise((resolve, reject) => {
      Sdk.Send("_rpcResolve"+_rpcToken+"(Tools.ReloadSkin());\n");
      rpcResolveHandlers.push(resolve);
    })
  },
};

if (typeof module !== 'undefined')
  module.exports = {Sdk:Sdk};
